## $Id: test_contract.py 12769 2015-03-15 13:14:08Z henrik $
##
## Copyright (C) 2014 Uli Fouquet & Henrik Bettermann
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##
"""
Tests for contracts.
"""
import decimal
from zope.interface.verify import verifyClass, verifyObject
from zope.component import createObject, getUtility, getUtilitiesFor
from zope.component.hooks import setSite
from hurry.workflow.interfaces import (
    IWorkflowInfo, IWorkflowState, InvalidTransitionError)
from waeup.ikoba.customers.interfaces import (
    IContractsContainer, IContract)
from waeup.ikoba.customers.contracts import (
    ContractsContainer, SampleContract, ContractPayer, ContractFinder,
    PayableContract,
    )
from waeup.ikoba.app import Company
from waeup.ikoba.customers.customer import Customer
from waeup.ikoba.payments.interfaces import (
    IPaymentItem, IPayer, IPayableFinder, IPayable,
    )
from waeup.ikoba.products.productoptions import ProductOption
from waeup.ikoba.testing import (FunctionalLayer, FunctionalTestCase)


class ContractsContainerTestCase(FunctionalTestCase):

    layer = FunctionalLayer

    def test_interfaces(self):
        # Make sure the correct interfaces are implemented.
        self.assertTrue(
            verifyClass(
                IContractsContainer, ContractsContainer)
            )
        self.assertTrue(
            verifyObject(
                IContractsContainer, ContractsContainer())
            )
        self.assertTrue(
            verifyClass(
                IContract, SampleContract)
            )
        self.assertTrue(
            verifyObject(
                IContract, SampleContract())
            )
        return

    def test_addContract(self):
        container = ContractsContainer()
        contract = createObject(u'waeup.SampleContract')
        id = contract.contract_id
        container.addContract(contract)
        self.assertEqual(container[id], contract)
        self.assertRaises(TypeError, container.addContract, object())
        self.assertEqual(contract.class_name, 'SampleContract')
        return

    def test_contract_workflow(self):
        contract = createObject(u'waeup.SampleContract')
        IWorkflowInfo(contract).fireTransition('create')
        self.assertEqual(IWorkflowState(contract).getState(), 'created')
        IWorkflowInfo(contract).fireTransition('submit')
        self.assertEqual(IWorkflowState(contract).getState(), 'submitted')
        IWorkflowInfo(contract).fireTransition('approve')
        self.assertEqual(IWorkflowState(contract).getState(), 'approved')
        IWorkflowInfo(contract).fireTransition('expire')
        self.assertEqual(IWorkflowState(contract).getState(), 'expired')
        IWorkflowInfo(contract).fireTransition('reset4')
        self.assertEqual(IWorkflowState(contract).getState(), 'created')
        self.assertRaises(InvalidTransitionError,
            IWorkflowInfo(contract).fireTransition, 'approve')
        IWorkflowInfo(contract).fireTransition('submit')
        IWorkflowInfo(contract).fireTransition('reset3')
        self.assertEqual(IWorkflowState(contract).getState(), 'created')
        return

    def test_contract_history(self):
        doc = createObject(u'waeup.SampleContract')
        IWorkflowInfo(doc).fireTransition('create')
        messages = ' '.join(doc.history.messages)
        self.assertTrue('Contract created by system' in messages)
        return


class TestContractHelpers(FunctionalTestCase):

    layer = FunctionalLayer

    def test_payer_adapter(self):
        # we can adapt IContract to IPayer (i.e. create a payer)
        customer = Customer()
        customer.firstname, customer.lastname = u'Anna', u'Tester'
        contract = createObject(u'waeup.SampleContract')
        customer['contracts'] = ContractsContainer()
        customer['contracts'].addContract(contract)
        result = IPayer(contract)
        self.assertTrue(isinstance(result, ContractPayer))
        verifyObject(IPayer, result)
        self.assertEqual(result.first_name, u'Anna')
        self.assertEqual(result.last_name, u'Tester')
        self.assertEqual(result.payer_id, customer.customer_id)

    def test_contract_finder_iface(self):
        # we have a contract finder that returns IPayableFinder data.
        verifyClass(IPayableFinder, ContractFinder)

    def test_contract_finder_registered(self):
        # the contract finder is a utility registered on startup
        util = getUtility(IPayableFinder, name='contracts_finder')
        self.assertTrue(isinstance(util, ContractFinder))
        utils = [util for name, util in getUtilitiesFor(IPayableFinder)
                 if isinstance(util, ContractFinder)]
        self.assertEqual(len(utils), 1)

    def create_contract_and_site(self):
        contract = SampleContract()
        option1 = ProductOption(u"Fee 1", decimal.Decimal("31.10"), "USD")
        option2 = ProductOption(u"Fee 2", decimal.Decimal("12.12"), "USD")
        contract.product_options = [option1, option2]
        contract.contract_id = u'CON1234'
        self.getRootFolder()['app'] = Company()
        app = self.getRootFolder()['app']
        setSite(app)
        return contract, app

    def test_contract_finder(self):
        # the contract finder can really find contracts
        contract, app = self.create_contract_and_site()
        app['mycontract'] = contract  # trigger cataloging
        finder = ContractFinder()
        result = finder.get_payable_by_id('CON1234')
        self.assertTrue(result is contract)

    def test_contract_finder_not_stored(self):
        # we get none if an id is not stored
        contract, app = self.create_contract_and_site()
        app['mycontract'] = contract  # trigger cataloging
        finder = ContractFinder()
        result = finder.get_payable_by_id('Not-a-valid-id')
        self.assertTrue(result is None)

    def test_contract_finder_no_catalog(self):
        # contract finder does not complain about missing catalog
        finder = ContractFinder()
        result = finder.get_payable_by_id('CON1234')
        self.assertTrue(result is None)


class TestContractAsPayable(FunctionalTestCase):

    layer = FunctionalLayer

    def test_adaptable(self):
        # we can turn contracts into payables.
        contract = SampleContract()
        payable = IPayable(contract)
        self.assertTrue(payable is not None)
        self.assertTrue(isinstance(payable, PayableContract))

    def test_payable_iface(self):
        # PayableContracts really provide IPayable
        contract = SampleContract()
        option1 = ProductOption(u"Fee 1", decimal.Decimal("31.10"), "USD")
        option2 = ProductOption(u"Fee 2", decimal.Decimal("12.12"), "USD")
        contract.product_options = [option1, option2]
        payable = PayableContract(contract)
        verifyObject(IPayable, payable)
        verifyClass(IPayable, PayableContract)

    def test_payable_simple_attributes(self):
        # the simple attribs are set correctly, according to context contract
        contract = SampleContract()
        contract.title = u'the title'
        option1 = ProductOption(u"Fee 1", decimal.Decimal("31.10"), "EUR")
        option2 = ProductOption(u"Fee 2", decimal.Decimal("12.12"), "EUR")
        contract.product_options = [option1, option2]
        payable = PayableContract(contract)
        self.assertTrue(contract.contract_id, payable.payable_id)
        self.assertEqual(payable.title, contract.title)
        self.assertEqual(payable.currency, 'EUR')

    def test_payable_items(self):
        # we can get payment items from payable
        contract = SampleContract()
        option1 = ProductOption(u"Fee 1", decimal.Decimal("31.10"), "EUR")
        option2 = ProductOption(u"Fee 2", decimal.Decimal("12.12"), "EUR")
        contract.product_options = [option1, option2]
        payable = PayableContract(contract)
        items = payable.payment_items
        self.assertTrue(isinstance(items, tuple))
        self.assertEqual(len(items), 2)
        verifyObject(IPaymentItem, items[0])
        verifyObject(IPaymentItem, items[1])
        self.assertEqual(items[0].item_id, '0')
        self.assertEqual(items[0].title, u'Fee 1')
        self.assertEqual(items[0].amount, decimal.Decimal("31.10"))

    def test_payable_no_items(self):
        # payables work also with no options set on contract
        contract = SampleContract()
        payable = PayableContract(contract)
        items = payable.payment_items
        self.assertTrue(isinstance(items, tuple))
        self.assertEqual(len(items), 0)
        self.assertEqual(payable.currency, None)
